import('./utils.js').then(utils => {
  import(/* @vite-ignore */utils.coreVersion()).then(core => {
    core.Sib.register({
      name: 'solid-notifications',
      use: [
        core.ListMixin,
        core.StoreMixin,
      ],
      attributes: {
        dateFormat: {
          type: String,
          default: 'DD/MM/YYYY - kk:mm'
        }
      },
      initialState: {
        favicon: null,
        title: null,
        unreadCount: null,
        unreadObjects: null,
        templates: null,
      },
      created() {
        import('https://cdn.skypack.dev/pin/favico.js-slevomat@v0.3.11-nyhOUAG4pt59qOhDEHAg/min/favico.js-slevomat.js').then(Favico => {
          this.favicon = new Favico.default({ animation: 'none' });
        });
        this.title = '';
        this.templates = {};
        this.unreadObjects = [];

        // read event -> mark notification as read
        window.addEventListener('read', event => {
          if (event.detail && event.detail.resource && event.detail.resource['@id']) {
            this._onReadResource(event.detail.resource['@id']);
          } else {
            this._printError('Resource parameter is missing. Check the documentation: https://git.startinblox.com/framework/sib-notifications');
          }
        })

        // click document -> close menu
        document.addEventListener("click", event => {
          if (!event.target.closest('.solid-notifications')) {
            this._closeMenu();
          }
        });

        // every 30 seconds -> refresh
        window.setInterval(() => {
          if (this.element.querySelector('details')  // prevent menu from closing on refresh
            && !this.element.querySelector('details').hasAttribute("open")) {
            this.refresh();
          }
        }, 30000);
      },

      empty() { },

      /**
       * Overwrite ListMixin method
       * @param {Object[]} resources - List of notifications
       */
      async renderDOM(resources) {
        const unreadNotifs = await this.getUnread(resources);
        const template = this.getTemplate(resources, unreadNotifs.length);
        core.render(template, this.element);
        this._updateUnread(unreadNotifs);
      },

      /**
       * Returns number of unread notifications
       * @param {Object[]} notifications - List of notifications
       *
       * @returns {number}
       */
      async getUnread(notifications) {
        const unreads = [];
        for (const n of notifications) {
          if (n && await n.unread && await n.object) {
            const obj = await Promise.all([n.type, n.object, n['object.@type']])
              .then(([type, object, objectType]) => ({ objectId: object['@id'], objectType, type }))
              .catch(() => null);
            unreads.push(obj);
          }
        }
        return unreads;
      },

      /* TEMPLATES */
      /**
       * Returns the template of the component
       * @param {Object[]} notifications - List of notifications
       * @param {number} unreadCount - count of unread notifications
       *
       * @returns {Object} returns a litHtml TemplateResult
       */
      getTemplate(notifications, unreadCount) {
        return core.html`
          <div class="solid-notifications">
            <details class="solid-notifications__container">
              <summary class="solid-notifications__button">
                <img src="${utils.path()}/assets/notification-icon.svg" />
                <div class="solid-notifications__counter">
                  <span data-nb-unread=${unreadCount}>${unreadCount}</span>
                </div>
              </summary>
              <div class="solid-notifications__arrow solid-notifications__triangle-back"></div>
              <div class="solid-notifications__arrow solid-notifications__triangle-shadow"></div>
              <div class="solid-notifications__list">
                ${core.until(this.getNotificationsTemplate(notifications))}
              </div>
            </details>
          </div>
        `;
      },

      /**
       * Filter notification and generate list template
       * @param {*} notifications
       * @returns
       */
      async getNotificationsTemplate(notifications) {
        const filteredNotifications = [];
        for (const n of notifications) {
          if (await n.object) {
            filteredNotifications.push(core.html`
            <div
              class="solid-notifications__item"
              @click=${() => this._onClickNotification(n.object)}
            >
              <div
                class="solid-notification"
                ?data-read=${core.until(
                  n.unread.then(u => !u), false
                )}
              >
                ${core.until(this.getNotificationTemplate(n))}
                <div class="solid-notification__read">
                  <button
                    @click=${(e) => this._onClickUnread(e, n)}
                    data-unread=${core.until(
                      n.unread.then(unread => unread ? 'unread' : 'read'), 'read'
                    )}
                  ></button>
                </div>
              </div>
            </div>`)
          }
        }
        return filteredNotifications;
      },

      /**
       * Generates notification template
       * @param {Object} notification - notification
       *
       * @returns {Object} returns a litHtml TemplateResult
       */
      async getNotificationTemplate(notification) {
        const object = await notification.object;
        const type = object['@type'];

        // Custom template
        if (type && this.templates[type]) {
          const data = await Promise.all([
            notification.summary,
            this._getFormattedDate(notification.date),
            notification.object,
            notification.author,
            notification.unread
          ]);
          let generatedTemplate;
          try {
            generatedTemplate = await core.Helpers.evalTemplateString(
              this.templates[type].innerHTML,
              {
                summary: this._sanitizeHTML(data[0]),
                date: data[1],
                object: data[2],
                author: data[3],
                unread: data[4],
              }
            );
          } catch (e) { generatedTemplate = '<div>Notification not available</div>'; }
          return core.html`${core.unsafeHTML(generatedTemplate)}`;
        }

        // Default template
        return core.html`
          <div class="solid-notification__avatar">
            <img src=${core.until(
              notification['author.account.picture'].then(img => img || ''), ''
            )} />
          </div>
          <div class="solid-notification__content">
            <solid-display
              class="solid-notification__title"
              data-src=${core.until(
                notification.object.then(r => r['@id']).catch(() => ''), ''
              )}
              fields="name"
            ></solid-display>
            <p class="solid-notification__summary">${core.until(notification.summary)}</p>
            <div class="solid-notification__date">
              ${core.until(this._getFormattedDate(notification.date), '')}
            </div>
          </div>
        `
      },

      /**
       * Refresh component data and re-render
       */
      async refresh() {
        if (this.resource) this.resource.clearCache(); // clear container cache
        core.store.getData(this.resourceId, this.context).then(this.updateDOM.bind(this));
      },

      /**
       * Saves a template used to render notifications related to specific objects
       * @param {string} type - rdf:type of the related object
       * @param {HTMLElement} template - template element to use for the notification
       */
      registerTemplate(type, template) {
        this.templates[type] = template;
      },

      /* PRIVATE */
      /**
       * Sends a request to mark a notification as read
       * @param {string} resourceId - id of the resource to patch
       * @private
       */
      async _onReadResource(resourceId) {
        if (!this.resource) return;
        let promises = [];

        for await(const notification of this.resource['ldp:contains']){
          if (await notification.unread && ((await notification['object.@id']) === resourceId || notification['@id'] === resourceId)) {
            promises.push(core.store.patch({
              "unread": false,
              "@context": this.context
            }, notification['@id']).catch(e => e));
          }
        }

        if (promises.length) {
          Promise.all(promises)
          .then(() => this.refresh())
          .catch((error) => this._printError(error));
        }
      },

      /**
       * Handler of the click on a notification
       * @param {Object} object - object of the notification
       * @private
       */
      async _onClickNotification(object) {
        object = await object;
        if (!object) return;
        window.dispatchEvent(new CustomEvent('requestNavigation', {
          detail: { resource: object }
        }));
        this._closeMenu();
      },

      /**
       * Handler of the click on an unread button
       * @param {Object} e - click event
       * @param {Object} resource - related resource
       * @private
       */
      _onClickUnread(e, resource) {
        e.stopPropagation();
        this._closeMenu();
        window.dispatchEvent(new CustomEvent('read', {
          detail: { resource: resource }
        }));
        e.target.disabled = true;
      },

      /**
       * Returns the date formatted as requested by the user
       * @param {Date} date - date to format
       * @private
       */
      _getFormattedDate(date) {
        return date.then(date => date ? moment((date).toString()).format(this.dateFormat) : '');
      },

      /**
       * Update the title and favicon with the new unread count
       * @param {array} unreadNotifs - objects unread
       * @private
       */
      _updateUnread(unreadNotifs) {
        if (unreadNotifs === null || (
          unreadNotifs.length === this.unreadCount &&
          this.sameUnreads(this.unreadObjects, unreadNotifs)
        )) return; // no change

        this.unreadCount = unreadNotifs.length;

        // favicon
        if (this.favicon) this.favicon.badge(this.unreadCount);

        // title
        if (this.title === "") this.title = document.title;
        if (unreadNotifs.length > 0) document.title = "(" + this.unreadCount + ") " + this.title;
        else document.title = this.title;

        // objects
        this.unreadObjects = [...unreadNotifs];
        window.dispatchEvent(new CustomEvent('notificationsCountRefresh', { bubbles: true })); // dispatch event to badges
      },

      /**
       * Check if 2 unread arrays are identical
       * @param {*} o1
       * @param {*} o2
       */
      sameUnreads(o1, o2) {
        for (const [index, o] of o1.entries()) {
          if (o.objectId !== o2[index].objectId ||
            o.objectType !== o2[index].objectType ||
            o.type !== o2[index].type) return false;
        }
        return true;
      },

      /**
       * Close notifications menu
       * @private
       */
      _closeMenu() {
        const notificationsMenu = this.element.querySelector('details');
        if (notificationsMenu) notificationsMenu.removeAttribute('open')
      },

      /**
       * Prints an error in console
       * @param {Object} error
       * @private
       */
      _printError(error) {
        console.error(error);
      },

      /**
       * Sanitize HTML to insert
       * @param {*} content
       * @returns string
       */
      _sanitizeHTML(content) {
        if (!content) return '';
        const element = document.createElement('div');
        element.innerText = content;
        return element.innerHTML;
      }
    });
  });
});
